#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

from unittest import mock

import pytest

from ci_scripts.bump_version import bump_version_from_news_files, news_ext_to_version_type, get_current_version, main


@pytest.mark.parametrize(
    "files,expected_version",
    (
        ((pathlib.Path("blah.feature"), pathlib.Path("blah.misc"), pathlib.Path("blah.bugfix")), "0.1.0"),
        ((pathlib.Path("blah.major"), pathlib.Path("blah.feature"), pathlib.Path("blah.bugfix")), "1.0.0"),
        ((pathlib.Path("blah.doc"), pathlib.Path("blah.misc"), pathlib.Path("blah.bugfix")), "0.0.1"),
    ),
)
def test_bumps_version_according_to_news_file_extension(files, expected_version):
    ver = bump_version_from_news_files("0.0.0", files)

    assert ver == expected_version


@pytest.mark.parametrize(
    "files",
    (
        tuple(),
        (pathlib.Path("blah.doc"), pathlib.Path("blah.removal"), pathlib.Path("blah.misc")),
        (pathlib.Path("blah.md"),),
    ),
)
def test_raises_when_significant_news_file_not_found(files):
    with pytest.raises(ValueError):
        bump_version_from_news_files("0.0.0", files)


@pytest.mark.parametrize(
    "news_exts,expected",
    (((".feature", ".major", ".bugfix"), "major"), ((".feature", ".bugfix"), "minor"), ((".misc", ".bugfix"), "patch")),
)
def test_converts_news_ext_to_version_type(news_exts, expected):
    version_type = news_ext_to_version_type(news_exts)

    assert version_type == expected


@mock.patch("ci_scripts.bump_version.git.Repo")
def test_get_version_returns_version_from_git(mock_repo):
    mock_version = "0.0.0"
    mock_repo().__enter__().git.describe.return_value = mock_version

    version_str = get_current_version()

    assert version_str == mock_version


@mock.patch("ci_scripts.bump_version.set_log_level")
@mock.patch("ci_scripts.bump_version.bump_version_from_news_files")
@mock.patch("ci_scripts.bump_version.get_current_version")
def test_main_execution_context(get_current_version, bump_version_from_news_files, set_log_level, tmp_path):
    class Args:
        verbosity = 0
        news_file_directory = pathlib.Path(tmp_path, "fake", "news")

    fake_version = "0.0.0"
    fake_args = Args()
    fake_args.news_file_directory.mkdir(parents=True)
    get_current_version.return_value = fake_version

    ret = main(fake_args)

    set_log_level.assert_called_once_with(fake_args.verbosity)
    bump_version_from_news_files.assert_called_once_with(fake_version, list(fake_args.news_file_directory.iterdir()))
    assert ret == 0



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import plistlib
from unittest import TestCase, mock

from mbed_tools.devices._internal.darwin.system_profiler import (
    get_all_usb_devices_data,
    get_end_usb_devices_data,
)


class TestGetAllUSBDevicesData(TestCase):
    @mock.patch("mbed_tools.devices._internal.darwin.system_profiler.subprocess.check_output")
    def test_returns_devices_list_from_system_profiler_call(self, check_output):
        check_output.return_value = b"""<plist version="1.0">
            <array>
                <dict>
                    <key>_items</key>
                    <array>
                        <string>foo</string>
                        <string>bar</string>
                    </array>
                </dict>
                <dict>
                    <key>vendor_id</key>
                    <string>hat</string>
                </dict>
            </array>
        </plist>"""

        self.assertEqual(get_all_usb_devices_data(), [{"_items": ["foo", "bar"]}, {"vendor_id": "hat"}])


@mock.patch("mbed_tools.devices._internal.darwin.system_profiler.get_all_usb_devices_data")
class TestGetEndUSBDevicesData(TestCase):
    def test_identifies_flat_end_devices(self, get_all_usb_devices_data):
        plist = b"""<plist version="1.0">
            <array>
                <dict>
                    <key>_name</key>
                    <string>USB Receiver</string>
                    <key>vendor_id</key>
                    <string>0xc53f</string>
                </dict>
            </array>
        </plist>"""
        get_all_usb_devices_data.return_value = plistlib.loads(plist)

        self.assertEqual(get_end_usb_devices_data(), [{"_name": "USB Receiver", "vendor_id": "0xc53f"}])

    def test_identifies_nested_end_devices(self, get_all_usb_devices_data):
        plist = b"""<plist version="1.0">
            <array>
                <dict>
                    <key>_items</key>
                    <array>
                        <dict>
                            <key>_name</key>
                            <string>USB Mouse</string>
                            <key>vendor_id</key>
                            <string>1234</string>
                        </dict>
                    </array>
                </dict>
                <dict>
                    <key>_name</key>
                    <string>USB2.0 Hub</string>
                    <key>_items</key>
                    <array>
                        <dict>
                            <key>_name</key>
                            <string>USB Flash</string>
                            <key>vendor_id</key>
                            <string>5678</string>
                        </dict>
                    </array>
                </dict>
            </array>
        </plist>"""
        get_all_usb_devices_data.return_value = plistlib.loads(plist)

        result = get_end_usb_devices_data()

        self.assertIn({"_name": "USB Mouse", "vendor_id": "1234"}, result)
        self.assertIn({"_name": "USB Flash", "vendor_id": "5678"}, result)

    def test_does_not_list_usb_hubs_or_buses(self, get_all_usb_devices_data):
        plist = b"""<plist version="1.0">
            <array>
                <dict>
                    <key>_name</key>
                    <string>USB31Bus</string>
                    <key>_items</key>
                    <array>
                        <dict>
                            <key>_name</key>
                            <string>USB3.0 Hub</string>
                        </dict>
                    </array>
                </dict>
                <dict>
                    <key>_name</key>
                    <string>USB20Bus</string>
                </dict>
            </array>
        </plist>"""
        get_all_usb_devices_data.return_value = plistlib.loads(plist)

        self.assertEqual(get_end_usb_devices_data(), [])



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Package definition for PyPI."""
import os

from setuptools import setup, find_packages

PROJECT_SLUG = "mbed-tools"
SOURCE_DIR = "src/mbed_tools"

repository_dir = os.path.dirname(__file__)

# Use readme needed as long description in PyPI
with open(os.path.join(repository_dir, "README.md"), encoding="utf8") as fh:
    long_description = fh.read()

setup(
    author="Mbed team",
    author_email="support@mbed.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Embedded Systems",
    ],
    description="Command line interface for Mbed OS.",
    keywords="Arm Mbed OS MbedOS cli command line tools",
    include_package_data=True,
    install_requires=[
        "python-dotenv",
        "Click==7.1",
        "pdoc3",
        "GitPython",
        "tqdm",
        "tabulate",
        "dataclasses; python_version<'3.7'",
        "requests>=2.20",
        "pywin32; platform_system=='Windows'",
        "psutil; platform_system=='Linux'",
        "pyudev; platform_system=='Linux'",
        "typing-extensions",
        "Jinja2",
        "pyserial",
        # Mbed OS 6.7 dependencies
        # TODO Remove these as part of fixing
        # https://github.com/ARMmbed/mbed-tools/issues/171
        "prettytable",
        "intelhex",
        "future",
    ],
    license="Apache 2.0",
    long_description_content_type="text/markdown",
    long_description=long_description,
    name=PROJECT_SLUG,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6,<4",
    url=f"https://github.com/ARMmbed/{PROJECT_SLUG}",
    entry_points={
        "console_scripts": [
            "mbedtools=mbed_tools.cli.main:cli",
            "mbed-tools=mbed_tools.cli.main:cli",
            "mbed_tools=mbed_tools.cli.main:cli",
        ]
    },
)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import json

import pytest

from mbed_tools.project import MbedProgram
from mbed_tools.build import generate_config
from mbed_tools.build.config import CMAKE_CONFIG_FILE
from mbed_tools.lib.exceptions import ToolsError


TARGETS = ["K64F", "NUCLEO_F401RE"]

TARGET_DATA = {
    "bootloader_supported": True,
    "c_lib": "std",
    "components": ["FLASHIAP", "SD"],
    "config": {
        "xip-enable": {
            "help": "Enable Execute In Place (XIP) on this "
            "target. Value is only significant if the "
            "board has executable external storage such "
            "as QSPIF. If this is enabled, customize "
            "the linker file to choose what text "
            "segments are placed on external storage",
            "value": False,
        }
    },
    "core": "Cortex-M4F",
    "default_toolchain": "ARM",
    "detect_code": ["0240"],
    "device_has": ["TRNG"],
    "device_name": "MK64FN1M0xxx12",
    "extra_labels": ["FRDM", "Freescale"],
    "features": ["PSA"],
    "is_disk_virtual": True,
    "labels": ["CORTEX", "CORTEX_M"],
    "macros": ["CPU_MK64FN1M0VMD12", "FSL_RTOS_MBED", "MBED_SPLIT_HEAP", "MBED_TICKLESS"],
    "printf_lib": "minimal-printf",
    "release_versions": ["5"],
    "static_memory_defines": True,
    "supported_application_profiles": ["full", "bare-metal"],
    "supported_c_libs": {"arm": ["std", "small"], "gcc_arm": ["std", "small"], "iar": ["std"]},
    "supported_form_factors": ["ARDUINO"],
    "supported_toolchains": ["ARM", "GCC_ARM", "IAR"],
    "trustzone": False,
}


def create_mbed_lib_json(lib_json_path, name, **kwargs):
    lib_json_path.parent.mkdir(parents=True, exist_ok=True)
    lib_json_path.write_text(json.dumps({"name": name, **kwargs}))


def create_mbed_app_json(root, **kwargs):
    (root / "mbed_app.json").write_text(json.dumps(kwargs))


@pytest.fixture
def program(tmp_path):
    prog = MbedProgram.from_new(tmp_path / "test-prog")
    # Overwrite the default mbed_app.json so it doesn't mess with our test env
    prog.files.app_config_file.write_text(json.dumps({"": ""}))
    # Create program mbed-os directory and fake targets.json
    prog.mbed_os.root.mkdir(parents=True)
    prog.mbed_os.targets_json_file.parent.mkdir(exist_ok=True, parents=True)
    prog.mbed_os.targets_json_file.write_text(json.dumps({target: TARGET_DATA for target in TARGETS}))
    return prog


@pytest.fixture
def program_in_mbed_os_subdir(tmp_path):
    mbed_os_path = tmp_path / "mbed-os"
    targets_json = mbed_os_path / "targets" / "targets.json"
    program_root = mbed_os_path / "test-prog"
    build_subdir = program_root / "__build" / "k64f"
    program_root.mkdir(parents=True, exist_ok=True)
    # Create program mbed-os directory and fake targets.json
    targets_json.parent.mkdir(exist_ok=True, parents=True)
    targets_json.write_text(json.dumps({target: TARGET_DATA for target in TARGETS}))
    return MbedProgram.from_existing(program_root, build_subdir, mbed_os_path=mbed_os_path)


@pytest.fixture(
    params=[(TARGETS[0], TARGETS[0]), (TARGETS[1], TARGETS[1]), (TARGETS[0], "*")],
    ids=lambda fixture_val: f"target: {fixture_val[0]}, filter: {fixture_val[1]}",
)
def matching_target_and_filter(request):
    return request.param


def test_target_and_toolchain_collected(program):
    target = "K64F"
    toolchain = "GCC_ARM"

    generate_config(target, toolchain, program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert target in config_text
    assert toolchain in config_text


def test_custom_targets_data_found(program):
    program.files.custom_targets_json.write_text(json.dumps({"IMAGINARYBOARD": TARGET_DATA}))
    target = "IMAGINARYBOARD"
    toolchain = "GCC_ARM"

    generate_config(target, toolchain, program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert target in config_text


def test_raises_error_when_attempting_to_customize_existing_target(program):
    program.files.custom_targets_json.write_text(json.dumps({TARGETS[0]: TARGET_DATA}))
    target = TARGETS[0]
    toolchain = "GCC_ARM"

    with pytest.raises(ToolsError):
        generate_config(target, toolchain, program)


def test_config_param_from_lib_processed_with_default_name_mangling(program):
    create_mbed_lib_json(
        program.mbed_os.root / "platform" / "mbed_lib.json",
        "platform",
        config={
            "stdio-convert-newlines": {
                "help": "Enable conversion to standard newlines on stdin/stdout/stderr",
                "value": True,
            }
        },
    )

    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_PLATFORM_STDIO_CONVERT_NEWLINES" in config_text


def test_config_param_from_lib_processed_with_user_set_name(program):
    create_mbed_lib_json(
        program.mbed_os.root / "platform" / "mbed_lib.json",
        "platform",
        config={
            "stdio-convert-newlines": {
                "help": "Enable conversion to standard newlines on stdin/stdout/stderr",
                "value": True,
                "macro_name": "ENABLE_NEWLINES",
            }
        },
    )

    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "ENABLE_NEWLINES" in config_text


def test_config_param_from_app_processed_with_default_name_mangling(program):
    create_mbed_app_json(
        program.root,
        config={
            "stdio-convert-newlines": {
                "help": "Enable conversion to standard newlines on stdin/stdout/stderr",
                "value": True,
            }
        },
    )

    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_APP_STDIO_CONVERT_NEWLINES" in config_text


def test_config_param_from_target_processed_with_default_name_mangling(program):
    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_TARGET_XIP_ENABLE=0" in config_text


@pytest.mark.parametrize(
    "macros",
    [["NS_USE_EXTERNAL_MBED_TLS"], ["NS_USE_EXTERNAL_MBED_TLS", "MBED_ENABLE_ERROR"]],
    ids=["single", "multiple"],
)
def test_macros_from_lib_collected(macros, program):
    create_mbed_lib_json(program.mbed_os.root / "connectivity" / "mbed_lib.json", "nanostack", macros=macros)

    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    for macro in macros:
        assert macro in config_text


@pytest.mark.parametrize(
    "macros",
    [["NS_USE_EXTERNAL_MBED_TLS"], ["NS_USE_EXTERNAL_MBED_TLS", "MBED_ENABLE_ERROR"]],
    ids=["single", "multiple"],
)
def test_macros_from_app_collected(macros, program):
    create_mbed_app_json(program.root, macros=macros)

    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    for macro in macros:
        assert macro in config_text


def test_macros_from_target_collected(program):
    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    for macro in TARGET_DATA["macros"]:
        assert macro in config_text


def test_target_labels_collected_as_defines(program):
    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    for label in TARGET_DATA["labels"] + TARGET_DATA["extra_labels"]:
        assert f"TARGET_{label}" in config_text

    for feature in TARGET_DATA["features"]:
        assert f"FEATURE_{feature}=1" in config_text

    for component in TARGET_DATA["components"]:
        assert f"COMPONENT_{component}=1" in config_text

    for device in TARGET_DATA["device_has"]:
        assert f"DEVICE_{device}=1" in config_text

    for form_factor in TARGET_DATA["supported_form_factors"]:
        assert f"TARGET_FF_{form_factor}" in config_text


def test_overrides_lib_config_param_from_app(matching_target_and_filter, program):
    target, target_filter = matching_target_and_filter
    create_mbed_lib_json(
        program.mbed_os.root / "mbed_lib.json", "platform", config={"stdio-baud-rate": {"value": 9600}},
    )

    create_mbed_app_json(program.root, target_overrides={target_filter: {"platform.stdio-baud-rate": 115200}})
    generate_config(target, "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_PLATFORM_STDIO_BAUD_RATE=115200" in config_text


def test_overrides_target_config_param_from_app(matching_target_and_filter, program):
    target, target_filter = matching_target_and_filter
    create_mbed_app_json(program.root, target_overrides={target_filter: {"target.xip-enable": True}})

    generate_config(target, "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_TARGET_XIP_ENABLE=1" in config_text


@pytest.mark.parametrize(
    "config_param, config_param_value, expected_output",
    [
        ("target.c_lib", "super", 'MBED_C_LIB "super"'),
        ("target.printf_lib", "maximal-printf", 'MBED_PRINTF_LIB "maximal-printf"'),
        ("target.extra_labels", ["NEW_LABELS"], "TARGET_NEW_LABELS"),
        ("target.supported_form_factors", ["BEAGLEBONE"], "TARGET_FF_BEAGLEBONE"),
        ("target.components", ["WARP_DRIVE"], "COMPONENT_WARP_DRIVE"),
        ("target.macros", ["DEFINE"], "DEFINE"),
        ("target.device_has", ["NOTHING"], "DEVICE_NOTHING"),
        ("target.features", ["ELECTRICITY"], "FEATURE_ELECTRICITY"),
    ],
)
def test_overrides_target_non_config_params_from_app(
    matching_target_and_filter, config_param, config_param_value, expected_output, program
):
    target, target_filter = matching_target_and_filter
    create_mbed_app_json(program.root, target_overrides={target_filter: {config_param: config_param_value}})

    generate_config(target, "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert expected_output in config_text


def test_overrides_target_config_param_from_lib(matching_target_and_filter, program):
    target, target_filter = matching_target_and_filter
    create_mbed_lib_json(
        program.root / "platform" / "mbed_lib.json",
        "platform",
        target_overrides={target_filter: {"target.xip-enable": True}},
    )

    generate_config(target, "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_TARGET_XIP_ENABLE=1" in config_text


def test_overrides_lib_config_param_from_same_lib(matching_target_and_filter, program):
    target, target_filter = matching_target_and_filter
    create_mbed_lib_json(
        program.mbed_os.root / "mbed_lib.json",
        "platform",
        config={"stdio-baud-rate": {"value": 9600}},
        target_overrides={target_filter: {"stdio-baud-rate": 115200}},
    )

    generate_config(target, "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_PLATFORM_STDIO_BAUD_RATE=115200" in config_text


def test_raises_when_attempting_to_override_lib_config_param_from_other_lib(matching_target_and_filter, program):
    target, target_filter = matching_target_and_filter
    create_mbed_lib_json(
        program.mbed_os.root / "platform" / "mbed_lib.json", "platform", config={"stdio-baud-rate": {"value": 9600}},
    )
    create_mbed_lib_json(
        program.mbed_os.root / "filesystem" / "mbed_lib.json",
        "filesystem",
        target_overrides={target_filter: {"platform.stdio-baud-rate": 115200}},
    )

    with pytest.raises(ToolsError):
        generate_config(target, "GCC_ARM", program)


@pytest.mark.parametrize(
    "config_param, config_param_value, expected_output",
    [
        ("target.macros_add", ["ENABLE_BOBBINS"], TARGET_DATA["macros"] + ["ENABLE_BOBBINS"]),
        ("target.extra_labels_add", ["EXTRA_LABEL_BOBBINS"], TARGET_DATA["extra_labels"] + ["EXTRA_LABEL_BOBBINS"]),
        ("target.features_add", ["FEATURE_BOBBINS"], TARGET_DATA["features"] + ["FEATURE_BOBBINS"]),
        ("target.components_add", ["COMPONENT_BOBBINS"], TARGET_DATA["components"] + ["COMPONENT_BOBBINS"]),
    ],
)
def test_target_list_params_can_be_added_to(
    matching_target_and_filter, config_param, config_param_value, expected_output, program
):
    target, target_filter = matching_target_and_filter
    create_mbed_app_json(
        program.root, target_overrides={target_filter: {config_param: config_param_value}},
    )

    generate_config(target, "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    for expected in expected_output:
        assert expected in config_text


@pytest.mark.parametrize(
    "config_param, config_param_value, expected_output",
    [
        ("target.macros_remove", [TARGET_DATA["macros"][0]], TARGET_DATA["macros"][0]),
        ("target.extra_labels_remove", [TARGET_DATA["extra_labels"][0]], TARGET_DATA["extra_labels"][0]),
        ("target.features_remove", [TARGET_DATA["features"][0]], TARGET_DATA["features"][0]),
        ("target.components_remove", [TARGET_DATA["components"][0]], TARGET_DATA["components"][0]),
    ],
)
def test_target_list_params_can_be_removed(
    matching_target_and_filter, config_param, config_param_value, expected_output, program
):
    target, target_filter = matching_target_and_filter
    create_mbed_app_json(
        program.root, target_overrides={target_filter: {config_param: config_param_value}},
    )

    generate_config(target, "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert expected_output not in config_text


def test_warns_when_attempting_to_override_nonexistent_param(matching_target_and_filter, program, caplog):
    target, target_filter = matching_target_and_filter
    override_name = "target.some-nonexistent-config-param"
    create_mbed_app_json(program.root, target_overrides={target_filter: {override_name: 999999}})

    generate_config(target, "GCC_ARM", program)

    assert override_name in caplog.text


def test_settings_from_multiple_libs_included(matching_target_and_filter, program):
    target, target_filter = matching_target_and_filter
    create_mbed_lib_json(
        program.mbed_os.root / "mbed_lib.json", "platform", config={"stdio-baud-rate": {"value": 9600}},
    )
    create_mbed_lib_json(
        program.mbed_os.root / "storage" / "mbed_lib.json",
        "filesystem",
        config={"read_size": {"macro_name": "MBED_LFS_READ_SIZE", "value": 64}},
    )

    generate_config(target, "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_PLATFORM_STDIO_BAUD_RATE=9600" in config_text
    assert "MBED_LFS_READ_SIZE=64" in config_text


def test_requires_config_option(program):
    create_mbed_app_json(program.root, requires=["bare-metal"])
    create_mbed_lib_json(program.mbed_os.root / "bare-metal" / "mbed_lib.json", "bare-metal", requires=["platform"])
    create_mbed_lib_json(
        program.mbed_os.root / "platform" / "mbed_lib.json", "platform", config={"stdio-baud-rate": {"value": 9600}},
    )
    create_mbed_lib_json(
        program.mbed_os.root / "storage" / "mbed_lib.json",
        "filesystem",
        config={"read_size": {"macro_name": "MBED_LFS_READ_SIZE", "value": 64}},
    )

    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_PLATFORM_STDIO_BAUD_RATE=9600" in config_text
    assert "MBED_LFS_READ_SIZE=64" not in config_text


def test_target_requires_config_option(program):
    create_mbed_app_json(program.root, target_overrides={"*": {"target.requires": ["ble"]}})
    create_mbed_lib_json(
        program.mbed_os.root / "bare-metal" / "mbed_lib.json",
        "ble",
        target_overrides={"*": {"target.requires": ["platform"]}},
    )
    create_mbed_lib_json(
        program.mbed_os.root / "platform" / "mbed_lib.json", "platform", config={"stdio-baud-rate": {"value": 9600}},
    )
    create_mbed_lib_json(
        program.mbed_os.root / "storage" / "mbed_lib.json",
        "filesystem",
        config={"read_size": {"macro_name": "MBED_LFS_READ_SIZE", "value": 64}},
    )

    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_PLATFORM_STDIO_BAUD_RATE=9600" in config_text
    assert "MBED_LFS_READ_SIZE=64" not in config_text


def test_config_parsed_when_mbed_os_outside_project_root(program_in_mbed_os_subdir, matching_target_and_filter):
    program = program_in_mbed_os_subdir
    target, target_filter = matching_target_and_filter
    create_mbed_lib_json(
        program.mbed_os.root / "mbed_lib.json", "platform", config={"stdio-baud-rate": {"value": 9600}},
    )
    create_mbed_lib_json(
        program.mbed_os.root / "storage" / "mbed_lib.json",
        "filesystem",
        config={"read_size": {"macro_name": "MBED_LFS_READ_SIZE", "value": 64}},
    )

    generate_config("K64F", "GCC_ARM", program)

    config_text = (program.files.cmake_build_dir / CMAKE_CONFIG_FILE).read_text()

    assert "MBED_CONF_PLATFORM_STDIO_BAUD_RATE=9600" in config_text
    assert "MBED_LFS_READ_SIZE=64" in config_text



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import contextlib
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase
from typing import Iterable


from mbed_tools.build._internal.find_files import find_files, filter_files, MbedignoreFilter, LabelFilter, _find_files


@contextlib.contextmanager
def create_files(files: Iterable[Path]):
    with TemporaryDirectory() as temp_directory:
        temp_directory = Path(temp_directory)
        for file in files:
            file_path = temp_directory / file
            file_directory = file_path.parent
            file_directory.mkdir(parents=True, exist_ok=True)
            file_path.touch()
        yield temp_directory


class TestFindFiles(TestCase):
    def test_finds_files_by_name(self):
        matching_paths = [
            Path("file.txt"),
            Path("sub_directory", "file.txt"),
        ]
        excluded_paths = [
            Path("not_interested.txt"),
            Path("sub_directory", "not_interested.txt"),
        ]

        with create_files(matching_paths + excluded_paths) as directory:
            subject = find_files("file.txt", directory)

        self.assertEqual(len(subject), len(matching_paths))
        for path in matching_paths:
            self.assertIn(Path(directory, path), subject)

    def test_respects_mbedignore(self):
        matching_paths = [
            Path("file.txt"),
        ]
        excluded_paths = [
            Path("foo", "file.txt"),
            Path("bar", "file.txt"),
        ]
        with create_files(matching_paths + excluded_paths) as directory:
            Path(directory, ".mbedignore").write_text("foo/*")
            Path(directory, "bar", ".mbedignore").write_text("*")

            subject = find_files("file.txt", directory)

        self.assertEqual(len(subject), len(matching_paths))
        for path in matching_paths:
            self.assertIn(Path(directory, path), subject)

    def test_respects_legacy_filters(self):
        matching_paths = [
            Path("file.txt"),
        ]
        excluded_paths = [
            Path("TESTS", "file.txt"),
            Path("bar", "TEST_APPS" "file.txt"),
        ]
        with create_files(matching_paths + excluded_paths) as directory:
            subject = find_files("file.txt", directory)

        self.assertEqual(len(subject), len(matching_paths))
        for path in matching_paths:
            self.assertIn(Path(directory, path), subject)

    def test_finds_all_with_no_filters(self):
        matching_paths = [
            Path("file.txt"),
            Path("bar", "file.txt"),
        ]
        with create_files(matching_paths) as directory:
            subject = _find_files("file.txt", directory)

        self.assertEqual(len(subject), len(matching_paths))
        for path in matching_paths:
            self.assertIn(Path(directory, path), subject)


class TestFilterFiles(TestCase):
    def test_respects_given_filters(self):
        matching_paths = [
            Path("foo", "file.txt"),
            Path("hello", "world", "file.txt"),
        ]
        excluded_paths = [
            Path("bar", "file.txt"),
            Path("foo", "foo-bar", "file.txt"),
        ]

        def my_filter(path):
            return "bar" not in str(path)

        subject = filter_files((matching_paths + excluded_paths), [my_filter])

        self.assertEqual(len(subject), len(matching_paths))
        for path in matching_paths:
            self.assertIn(path, subject)


class TestLabelFilter(TestCase):
    def test_matches_paths_not_following_label_rules(self):
        subject = LabelFilter("TARGET", ["BAR", "BAZ"])

        self.assertFalse(subject(Path("mbed-os", "TARGET_FOO", "some_file.c")))
        self.assertFalse(subject(Path("mbed-os", "TARGET_BAR", "TARGET_FOO", "other_file.c")))

    def test_does_not_match_paths_following_label_rules(self):
        subject = LabelFilter("TARGET", ["BAR", "BAZ"])

        self.assertTrue(subject(Path("mbed-os", "TARGET_BAR", "some_file.c")))
        self.assertTrue(subject(Path("mbed-os", "COMPONENT_X", "header.h")))
        self.assertTrue(subject(Path("mbed-os", "COMPONENT_X", "TARGET_BAZ", "some_file.c")))
        self.assertTrue(subject(Path("README.md")))


class TestMbedignoreFilter(TestCase):
    def test_matches_files_by_name(self):
        subject = MbedignoreFilter(("*.py",))

        self.assertFalse(subject("file.py"))
        self.assertFalse(subject("nested/file.py"))
        self.assertTrue(subject("file.txt"))

    def test_matches_wildcards(self):
        subject = MbedignoreFilter(("*/test/*",))

        self.assertFalse(subject("foo/test/bar.txt"))
        self.assertFalse(subject("bar/test/other/file.py"))
        self.assertTrue(subject("file.txt"))

    def test_from_file(self):
        with TemporaryDirectory() as temp_directory:
            mbedignore = Path(temp_directory, ".mbedignore")
            mbedignore.write_text(
                """
# Comment

foo/*.txt
*.py
"""
            )

            subject = MbedignoreFilter.from_file(mbedignore)

            self.assertEqual(
                subject._patterns, (str(Path(temp_directory, "foo/*.txt")), str(Path(temp_directory, "*.py")),)
            )



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#

import tempfile
import pathlib
import json

from unittest import TestCase

from click.testing import CliRunner

from mbed_tools.cli.configure import configure

target_json = json.dumps(
    {
        "TARGET": {
            "core": None,
            "trustzone": False,
            "default_toolchain": "ARM",
            "supported_toolchains": None,
            "extra_labels": [],
            "supported_form_factors": [],
            "components": [],
            "is_disk_virtual": False,
            "macros": [],
            "device_has": [],
            "features": [],
            "detect_code": [],
            "c_lib": "std",
            "bootloader_supported": False,
            "static_memory_defines": True,
            "printf_lib": "minimal-printf",
            "supported_c_libs": {"arm": ["std"], "gcc_arm": ["std", "small"], "iar": ["std"]},
            "supported_application_profiles": ["full"],
        }
    }
)

mbed_app_json = json.dumps(
    {"target_overrides": {"*": {"target.c_lib": "small", "target.printf_lib": "minimal-printf"}}}
)


class TestConfigureRegression(TestCase):
    def test_generate_config_called_with_correct_arguments(self):
        with tempfile.TemporaryDirectory() as tmpDir:
            tmpDirPath = pathlib.Path(tmpDir)
            pathlib.Path(tmpDirPath / "mbed-os.lib").write_text("https://github.com/ARMmbed/mbed-os")
            pathlib.Path(tmpDirPath / "mbed_app.json").write_text(mbed_app_json)
            pathlib.Path(tmpDirPath / "mbed-os").mkdir()
            pathlib.Path(tmpDirPath / "mbed-os" / "targets").mkdir()
            pathlib.Path(tmpDirPath / "mbed-os" / "targets" / "targets.json").write_text(target_json)

            result = CliRunner().invoke(
                configure, ["-m", "Target", "-t", "gcc_arm", "-p", tmpDir], catch_exceptions=False
            )
            self.assertIn("mbed_config.cmake has been generated and written to", result.output)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib
import re

from unittest import TestCase, mock

from mbed_tools.targets import Board
from mbed_tools.targets.exceptions import MbedTargetsError

from tests.devices.factories import CandidateDeviceFactory
from mbed_tools.devices.device import Device
from mbed_tools.devices._internal.exceptions import NoBoardForCandidate

from mbed_tools.devices.devices import (
    get_connected_devices,
    find_connected_device,
    find_all_connected_devices,
)
from mbed_tools.devices.exceptions import DeviceLookupFailed, NoDevicesFound


@mock.patch("mbed_tools.devices.devices.detect_candidate_devices")
@mock.patch("mbed_tools.devices.devices.resolve_board")
class TestGetConnectedDevices(TestCase):
    def test_builds_devices_from_candidates(self, resolve_board, detect_candidate_devices):
        candidate = CandidateDeviceFactory()
        detect_candidate_devices.return_value = [candidate]

        connected_devices = get_connected_devices()
        self.assertEqual(
            connected_devices.identified_devices,
            [
                Device(
                    serial_port=candidate.serial_port,
                    serial_number=candidate.serial_number,
                    mount_points=candidate.mount_points,
                    mbed_board=resolve_board.return_value,
                )
            ],
        )
        self.assertEqual(connected_devices.unidentified_devices, [])
        resolve_board.assert_called_once_with(candidate)

    @mock.patch.object(Board, "from_offline_board_entry")
    def test_skips_candidates_without_a_board(self, board, resolve_board, detect_candidate_devices):
        candidate = CandidateDeviceFactory()
        resolve_board.side_effect = NoBoardForCandidate
        detect_candidate_devices.return_value = [candidate]
        board.return_value = None

        connected_devices = get_connected_devices()
        self.assertEqual(connected_devices.identified_devices, [])
        self.assertEqual(
            connected_devices.unidentified_devices,
            [
                Device(
                    serial_port=candidate.serial_port,
                    serial_number=candidate.serial_number,
                    mount_points=candidate.mount_points,
                    mbed_board=None,
                )
            ],
        )

    def test_raises_device_lookup_failed_on_internal_error(self, resolve_board, detect_candidate_devices):
        resolve_board.side_effect = MbedTargetsError
        detect_candidate_devices.return_value = [CandidateDeviceFactory()]

        with self.assertRaises(DeviceLookupFailed):
            get_connected_devices()


@mock.patch("mbed_tools.devices.devices.find_all_connected_devices")
class TestFindConnectedDevice(TestCase):
    def test_finds_device_with_matching_name(self, mock_find_connected_devices):
        target_name = "K64F"
        mock_find_connected_devices.return_value = [
            mock.Mock(mbed_board=mock.Mock(board_type=target_name, spec=True), serial_number="123", spec=True)
        ]

        dev = find_connected_device(target_name)

        self.assertEqual(target_name, dev.mbed_board.board_type)

    def test_finds_device_with_matching_name_identifier(self, mock_find_connected_devices):
        target_name = "K64F"
        target_identifier = 1
        mock_find_connected_devices.return_value = [
            mock.Mock(mbed_board=mock.Mock(board_type=target_name, spec=True), serial_number="123", spec=True),
            mock.Mock(mbed_board=mock.Mock(board_type=target_name, spec=True), serial_number="456", spec=True),
        ]

        dev = find_connected_device(target_name, target_identifier)

        self.assertEqual(dev.serial_number, "456")

    def test_raises_when_multiple_matching_name_no_identifier(self, mock_find_connected_devices):
        target_name = "K64F"
        mock_find_connected_devices.return_value = [
            mock.Mock(
                serial_port="tty.0",
                mount_points=[pathlib.Path("/board")],
                mbed_board=mock.Mock(board_type=target_name, spec=True),
                serial_number="456",
                spec=True,
            ),
            mock.Mock(
                serial_port="tty.1",
                mount_points=[pathlib.Path("/board2")],
                mbed_board=mock.Mock(board_type=target_name, spec=True),
                serial_number="123",
                spec=True,
            ),
        ]

        with self.assertRaises(DeviceLookupFailed) as ex:
            find_connected_device("K64F", None)
        self.assertTrue("Multiple" in ex.exception.args[0])

    def test_raises_when_identifier_out_of_bounds(self, mock_find_connected_devices):
        target_name = "K64F"
        mock_find_connected_devices.return_value = [
            mock.Mock(
                serial_port="tty.0",
                mount_points=[pathlib.Path("/board")],
                mbed_board=mock.Mock(board_type=target_name, spec=True),
                serial_number="456",
                spec=True,
            ),
            mock.Mock(
                serial_port="tty.1",
                mount_points=[pathlib.Path("/board2")],
                mbed_board=mock.Mock(board_type=target_name, spec=True),
                serial_number="123",
                spec=True,
            ),
        ]

        with self.assertRaises(DeviceLookupFailed) as ex:
            find_connected_device("K64F", 2)
        self.assertTrue("valid" in ex.exception.args[0])


@mock.patch("mbed_tools.devices.devices.get_connected_devices")
class TestFindAllConnectedDevices(TestCase):
    def test_finds_all_devices_with_matching_name(self, mock_get_connected_devices):
        target_name = "K64F"
        mock_get_connected_devices.return_value = mock.Mock(
            identified_devices=[
                mock.Mock(mbed_board=mock.Mock(board_type=target_name, spec=True), serial_number="456", spec=True),
                mock.Mock(mbed_board=mock.Mock(board_type=target_name, spec=True), serial_number="123", spec=True),
                mock.Mock(mbed_board=mock.Mock(board_type="DISCO", spec=True), serial_number="345", spec=True),
            ],
            spec=True,
        )

        devices = find_all_connected_devices(target_name)

        self.assertEqual(len(devices), 2)
        self.assertEqual(devices[0].serial_number, "123")
        self.assertEqual(devices[1].serial_number, "456")

    def test_raises_when_no_mbed_enabled_devices_found(self, mock_get_connected_devices):
        mock_get_connected_devices.return_value = mock.Mock(identified_devices=[], spec=True)

        with self.assertRaises(NoDevicesFound):
            find_all_connected_devices("K64F")

    def test_raises_when_device_matching_target_name_not_found(self, mock_get_connected_devices):
        target_name = "K64F"
        connected_target_name = "DISCO"
        connected_target_serial_port = "tty.BLEH"
        connected_target_mount_point = [pathlib.Path("/dap")]
        mock_get_connected_devices.return_value = mock.Mock(
            identified_devices=[
                mock.Mock(
                    serial_port=connected_target_serial_port,
                    mount_points=connected_target_mount_point,
                    mbed_board=mock.Mock(board_type=connected_target_name, spec=True),
                    spec=True,
                )
            ],
            spec=True,
        )

        with self.assertRaisesRegex(
            DeviceLookupFailed,
            f".*(target: {re.escape(connected_target_name)}).*(port: {re.escape(connected_target_serial_port)}).*"
            f"(mount point.*: {re.escape(str(connected_target_mount_point))})",
        ):
            find_all_connected_devices(target_name)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase, mock

from mbed_tools.project._internal.progress import ProgressReporter, ProgressBar


class TestProgressBar(TestCase):
    @mock.patch("mbed_tools.project._internal.progress.ProgressBar.update")
    def test_updates_progress_bar_with_correct_block_size(self, mock_bar_update):
        bar = ProgressBar(total=100)
        bar.update_progress(1, 1)

        mock_bar_update.assert_called_once_with(1)

    def test_sets_total_attribute_to_value_of_total_size(self):
        bar = ProgressBar()

        self.assertIsNone(bar.total)

        bar.update_progress(1, 2, total_size=33)

        self.assertEqual(bar.total, 33)


@mock.patch("mbed_tools.project._internal.progress.ProgressBar", autospec=True)
class TestProgressReporter(TestCase):
    def test_creates_progress_bar_on_begin_opcode(self, mock_progress_bar):
        reporter = ProgressReporter()
        reporter._cur_line = "begin"
        reporter.update(reporter.BEGIN, 1)

        mock_progress_bar.assert_called_once()

    def test_closes_progress_bar_on_end_opcode(self, mock_progress_bar):
        reporter = ProgressReporter()
        reporter.bar = mock_progress_bar()
        reporter.update(reporter.END, 1)

        reporter.bar.close.assert_called_once()
        reporter.bar.update.assert_not_called()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pytest

from mbed_tools.lib.python_helpers import minimum_python_version, named_tuple_with_defaults, flatten_nested


class TestPythonHelpers:
    def test_python_version(self):
        # Tools only support Python>= 3.6
        assert minimum_python_version(3, 6)
        # With a bit of luck, python 100.100 is not coming tomorrow
        assert not minimum_python_version(100, 100)

    def test_named_tuple_with_defaults(self):
        named_tuple = named_tuple_with_defaults("TestNamedTuple", field_names=["field1", "field2"], defaults=[1, 2])
        a_tuple = named_tuple()
        assert a_tuple
        assert a_tuple.field1 == 1
        assert a_tuple.field2 == 2

        a_tuple = named_tuple(field1=15)
        assert a_tuple.field1 == 15
        assert a_tuple.field2 == 2


@pytest.mark.parametrize(
    "input_list, expected_result",
    (
        ([1, 2, 3, 4, 5], [1, 2, 3, 4, 5]),
        ([1, [2], 3, 4, 5], [1, 2, 3, 4, 5]),
        ([[1, 2, 3], [4, 5]], [1, 2, 3, 4, 5]),
        ([(1, 2, 3), (4, 5)], [1, 2, 3, 4, 5]),
        ([[1, [[2, [3]]]], [4, 5]], [1, 2, 3, 4, 5]),
        ([["alan", [["bob", ["sally"]]]], ["jim", "jenny"]], ["alan", "bob", "sally", "jim", "jenny"]),
    ),
)
def test_flatten_nested_list(input_list, expected_result):
    assert flatten_nested(input_list) == expected_result
